/* 
Reziproke Frequenzmessung mit STM32H750 mit 8-stelliger Auflsung.
Eingangsfrequenbereich: 0,1 Hz - 100 MHz

Die Frequenz wird aus den Eingangsimpulsen/Zeit errechnet. Dazu werden die 
Timer15 und Timer16 verwendet. Als Zeitbasis fr alle Messungen dient die
interne Frequenz SystemTimerClock mit (hier) 240 MHz. 

Timer15 zhlt am Eingang TIM15_CH1 (PE5) forlaufend die Eingangsimpulse. 
Timer16 zhlt fortlaufend die interen Impulse von SystemTimerClock (240 MHz).
Beide Zhler werden einmalig gestartet und dann weder gestoppt noch gelscht.
Timer6 erzeugt das Timing fr die Messdauer im 1 ms Raster.

Ein ext. D-FF erzeugt auf Anforderung mit Signal TRIG2 (PC0) synchron zu einem 
Eingangsimpuls das Signal CAP2 (TIM15_CH2 (PE6) fr Timer15 und TIM16_CH1 (PB8) 
fr Timer16). In den zugehrigen Capture-Registern werden die Anzahl der Ereignisse
sowie deren genauer Zeitpunkt festgehalten. Zustzlich werden auch die berlufe
der Timer bercksichtigt, soda beide Messwerte als uint32_t Variablen zur
Auswertung bereitstehen: "F2_end_ereignis" und "F2_end_zeit".
Diese Zeitpunkte werden in der ISR TIM15_IRQHandler ermittelt.

Ein einzelner Messpunkt liefert noch kein Ergebnis, sondern es mssen ein 
Startzeitpunkt und ein Endzeitpunkt vorliegen, um aus deren Differenz Ereignisse
und Zeit zu erhalten: "F2_mess_ereignisse" und "F2_mess_zeit".

"F2_mess_zeit" enthlt die Anzahl der internen Referenzimpulse und mu daher noch
mit Division durch SystemTimerClock auf Sekunde umgerechnet werden. Das Ergebnis
ist dann "eff_zeit2".
Das Messergebnis ergibt sich dann als: "F2_ergebnis = F2_mess_ereignisse / eff_zeit2".

Nach einer Auswertung werden die letzten Endpunkte zu den Startpunkten fr das
neue Intervall.

Wie geschrieben wird eine Auswertung immer mit dem Signal TRIG2 angefordert und
synchron zu einem Eingangsimpuls gestartet. Wenn man schnell viele Messwerte
braucht, knnte man nach jeder Auswertung mit TRIG2 gleich eine neue Auswertung
anfordern. Der hier verwendete C STM32H750 schafft so einige 100000 Messungen/s,
sofern man auf die ser. Datenausgabe verzichtet.
Bentigt man jedoch eine mglichst hohe Auflsung der Messwerte, mu man dafr
sorgen, da die Zeitmessung mglichst hoch aufgelst ist. Dies gilt nicht fr 
die Eingangsimpulse, da diese den Zeitpunkt der Auswertung synchron vorgeben.

Die Zeitmessung luft mit der internen Frequenz SystemTimerClock. Die hier 
eingestellte Maximalfrequenz betrgt 240 MHz oder anders geschrieben 2.40E08 Hz.
Am Exponenten ist zu erkennen, da in einer Sekunde eine 8-stellige Auflsung
bei der Zeitmessung vorliegt. Selbst bei 1/2 s Messzeit sind mit 1.20E08
Referenzimpulsen noch 8-stellige Werte erreichbar.
Fr ein 8-stelliges Messergebnis sind die Auswertungen mit einem Abstand von
>= 0,5 s anzufordern.
Verkrzt man diese Zeit auf z.B. 1 ms, geht die Auflsung auf 5 Stellen/Messwert
zurck. Da jede Auswertung synchron zu einem Eingangsimpuls gestartet wird,
wird die Messrate bei niedrigen Frequenzen reduziert. Bei 1 Hz Eingangssignal
ist nur 1 Messung/s mglich.
Fehlt das Eingangssignal ganz, wird per Timeout die Messung abgebrochen und
neu gestartet.

Die Messwerte werden ber USART3 ausgegeben: "sende_sting(s)"

2020-10-21
Michael Nowak
http://www.mino-elektronik.de
Alle Angaben ohne Gewaehr !
*/

#define F2_MESSUNG

#include <stdio.h>
#include <string.h>
#include "stm32h7xx.h"
extern void sende_string(char *s);    // Ausgabe ber USART3
extern void lcd_zeile1(char *s);
extern void lcd_zeile2(char *s);
extern uint8_t lcd_init(void);

extern uint32_t SystemTimerClock;

#define DEF_F2_TAKT       240000000   // 240 MHz
#define DEF_F2_MESSZEIT   666         // 0,666 s
#define DEF_TIMEOUT       10000       // nach 10 s
#define MAX_ZEICHEN       50          // fuer Ergebnis String

#define PRIO_TIMER        4           // hohe Prioritt fr T15 und T16
#define PRIO_TIM6         14          // niedrig, unkritisch
#define AFR_TIM15         4           // fr T15  lt. Datenblatt
#define AFR_TIM16         1           // fr T16 lt. Datenblatt


#define BIT(x) (1<<x)
enum gpio_modus {IN, OUT, AF, ANALOG};
enum gpio_speed {LOW_SPEED, MEDIUM_SPEED, HIGH_SPEED, V_HIGH_SPEED};
enum gpio_pullup {NO_PULL, PULL_UP, PULL_DOWN};

#define CLR_MODER         3         // Lschmaske
#define SET_GPIO          0x1       // untere Bits
#define RESET_GPIO        0x10000   // obere Bits

// PortB
#define TIM16_CAPTURE2    8
// PortC
#define TRIG2             0
// PortE
#define TIM15_ZAEHLER2    5
#define TIM15_CAPTURE2    6

enum ablauf {MESSEN=0, AUSLESEN, AUSWERTEN};        // die Phasen der Messung

double              F2_ergebnis,                    // in Hz
                    eff_zeit2;                      // eff. Messzeit in s
double              F2_ref_frequenz = DEF_F2_TAKT;  // Vorgabewerte
int32_t             F2_ref_offset = 0,              // fr Abgleich von F2_ref
                    F2_messzeit = DEF_F2_MESSZEIT,
                    F2_timeout = DEF_TIMEOUT;

volatile uint8_t    F2_messwert_vorhanden,          // gueltige Messungen
                    F2_mess_status;                 // Ablaufsteuerung
                                   
volatile uint32_t   F2_mess_dauer;                  // in ms
uint32_t            F2_start_ereignis,              // Impulse zu Beginn der Messung
                    F2_start_zeit,                  // rel. Zeit: Beginn der Messung
                    F2_zeit_ueberlauf,
                    F2_end_zeit,                    // Zeitpunkt der Auswertung
                    F2_mess_zeit,                   // genaue Zeit der Messung
                    F2_mess_ereignisse,             // Impulse der Messung
                    F2_ueberlauf; 
volatile uint32_t   F2_end_ereignis;                // Impulse am Ende der Messung


// direktes Setzen oder Ruecksetzen eines Port-Pins
void set_pin(GPIO_TypeDef* GPIOx, uint8_t number, char set) {
  if(set) {
    GPIOx->BSRR = (SET_GPIO << number);       // IO-Bit setzen
  } else {
    GPIOx->BSRR = (RESET_GPIO << number);     // IO-Bit loeschen
  }
}


// 1 kHz ISR mit T6 erzeugen
void init_t6(void)
{
static uint8_t init;
  if(!init) {
    init = 1;
    RCC->APB1LENR |= RCC_APB1LENR_TIM6EN;
    NVIC_EnableIRQ(TIM6_DAC_IRQn);
    NVIC_SetPriority(TIM6_DAC_IRQn,15);     // geringe Prioritaet
    TIM6->PSC = SystemTimerClock/1000000-1; // Vorteiler auf 1 MHz
    TIM6->ARR = 1000-1;                     // Teiler fr 1ms
    TIM6->CR1 = TIM_CR1_CEN;                // starten
    TIM6->DIER = TIM_DIER_UIE;              // und Interrupt freigeben
  }    
} 

void TIM6_DAC_IRQHandler(void)              // Aufruf mit 1 kHz
{
  if(TIM6->SR & TIM_SR_UIF) {               // nur bei berlufen von T16
    TIM6->SR = ~TIM_SR_UIF;
    F2_mess_dauer++;                        // Messzeit erfassen
    __ISB();                                // gelschtes Flag abwarten
  }
}
 

// Timer15 zhlt die ext. Eingangsimpulse: Ereignismessung
void init_t15(void)
{
uint32_t tmp;  
  RCC->APB2ENR |= RCC_APB2ENR_TIM15EN;
	RCC->AHB4ENR |= RCC_AHB4ENR_GPIOEEN;          // PortE aktivieren
  NVIC_EnableIRQ(TIM15_IRQn);                   // Int-T15 aktivieren
  NVIC_SetPriority(TIM15_IRQn,PRIO_TIMER);

  TIM15->CR1 = 0;                               // T15 stoppen, falls schon aktiv
  tmp = GPIOE->MODER;
  tmp &= ~(CLR_MODER << (TIM15_ZAEHLER2 * 2) | 
           CLR_MODER << (TIM15_CAPTURE2 * 2));  // MODER auf 0                
  tmp |= (AF << (TIM15_ZAEHLER2 * 2) |          // altern. Funktion
          AF << (TIM15_CAPTURE2 * 2));                       
  GPIOE->MODER = tmp;

// PE5 als Zhlereingang, PE6 als capture-input zu T15
  GPIOE->AFR[0] |= (AFR_TIM15 << (TIM15_ZAEHLER2 * 4) | 
                    AFR_TIM15 << (TIM15_CAPTURE2 * 4));                   

  TIM15->CCMR1 = TIM_CCMR1_CC2S_0;              // capture zuordnung  IC2 -> TI2
  TIM15->CCER = TIM_CCER_CC2E;                  // pos. Flanken an IC2 freigeben
  TIM15->SMCR |= TIM_SMCR_TS_2 + TIM_SMCR_TS_0; // ext. Signal von TI1
  TIM15->SMCR |= TIM_SMCR_SMS_2 + TIM_SMCR_SMS_1 + 
                 TIM_SMCR_SMS_0;                // Zhlereingang von TI1

  TIM15->DIER |= TIM_DIER_CC2IE + TIM_DIER_UIE; // mit Capture + berlauf interrupts
  TIM15->CR1 |= TIM_CR1_CEN;                    // T15 starten
}

// Timer16 zhlt die internen Referenzimpulse: Zeitmessung
void init_t16(void)
{
uint32_t tmp;
  RCC->APB2ENR |= RCC_APB2ENR_TIM16EN;
	RCC->AHB4ENR |= RCC_AHB4ENR_GPIOBEN;          // PortB aktivieren
  NVIC_EnableIRQ(TIM16_IRQn);                   // Int-T16 aktivieren
  NVIC_SetPriority(TIM16_IRQn,PRIO_TIMER);

  TIM16->CR1 = 0;                               // T16 stoppen, falls schon aktiv
  tmp = GPIOB->MODER;
  tmp &= ~(CLR_MODER << (TIM16_CAPTURE2 * 2));  // MODER PB8 = 0       
  tmp |= (2 << (TIM16_CAPTURE2 * 2));           // PB8  auf Ausgang        
  GPIOB->MODER = tmp;

// PB8 als capture-input zu T16
  GPIOB->AFR[1] |=  (AFR_TIM16 << ((TIM16_CAPTURE2-8) * 4));                   
  TIM16->CCMR1 =  TIM_CCMR1_CC1S_0;             // capture zuordnung  IC1 -> TI1
  TIM16->CCER = TIM_CCER_CC1E;                  // pos. Flanken an IC1 freigeben

  TIM16->DIER |= TIM_DIER_UIE;                  // mit interrupt bei berlauf
  TIM16->CR1 |= TIM_CR1_CEN;                    // T16 starten
}

// nur berlaufe zhlen
void TIM16_IRQHandler(void)
{
  if(TIM16->SR & TIM_SR_UIF) {                  // nur bei berlufen von T16
    TIM16->SR = ~TIM_SR_UIF;
    F2_zeit_ueberlauf += 0x10000;
  }
  __ISB();                                      // gelschtes Flag abwarten
}

// ISR zu Timer15
// synchron zu einem Eingangsimpuls Ereignisse und Zeitpunkt ermitteln
// anschlieend Auswertung freigeben
// ferner berlaufe bei der Ereigniszhlung zhlen
// Auswertung Einzel-/Gesamtmessung synchron zu Capture-Signal an PE6: D-FF wurde gesetzt
void TIM15_IRQHandler(void)
{
uint32_t temp;
  if(TIM15->SR & TIM_SR_CC2IF) {                        // bei Capture-Ereignis
    TIM15->SR = ~TIM_SR_CC2IF;                          // flag wieder loeschen
    temp = TIM15->CCR2;                                 // Capture-Wert Ereignisse
    F2_end_ereignis = F2_ueberlauf + temp;              // Endwert Ereignisse
    if((TIM15->SR & TIM_SR_UIF) && (temp < 0x8000))     // evtl. Ueberlauf T15 noch offen?
      F2_end_ereignis += 0x10000;                       // nur, wenn capture-int + overflow-int gleichzeitig !

    temp = TIM16->CCR1;                                 // Capture-Wert Zeit
    F2_end_zeit = F2_zeit_ueberlauf + temp;             // Endwert Zeit
    if((TIM16->SR & TIM_SR_UIF) && (temp < 0x8000)) {   // evtl. Ueberlauf T16 noch offen?
      F2_end_zeit += 0x10000;                           // dann korrigieren
    }  
    set_pin(GPIOC, TRIG2, 0);                           // Trigger-Ausgang wieder auf '0'
    F2_mess_status = AUSWERTEN;                         // Daten fertig fuer Auswertung
  }
  if(TIM15->SR & TIM_SR_UIF) {                          // bei berlufen von T15
     TIM15->SR = ~TIM_SR_UIF;                           // flag wieder loeschen
     F2_ueberlauf += 0x10000;                           // und berlufe zhlen
  }
  __ISB();                                              // gelschte Flags abwarten
}


/* direkte Messung mit internen Timern mit max. 8 Stellen/s
Timer15 zhlt die externen Impulse
Timer16 mit die Zeit mit internem Takt
*/
void F2_messung(uint8_t init_timer)
{
static uint8_t  init = 0;
static char s[MAX_ZEICHEN];
  if(init_timer) {
    lcd_init();
    init = 0;                                   // Anzeige + Messung initialisieren
    init_t6();                                  // T6 fr Messdauer in 1 ms Schritten
    init_t15();                                 // T15 als Ereignisszhler mit capture-modus starten
    init_t16();                                 // T16 als Referenz-Zhler
    RCC->AHB4ENR |= RCC_AHB4ENR_GPIOCEN;        // PortC aktivieren
    GPIOC->MODER    &= ~(CLR_MODER << TRIG2*2); // neutral und dann
    GPIOC->MODER    |= OUT << TRIG2*2;          // als Trigger-Ausgang
    GPIOC->OSPEEDR  |= V_HIGH_SPEED << TRIG2*2; // mit steilen Flanken
  }
    if(!init) {
      F2_mess_status  = MESSEN;                 // mit MESSEN Beginnen
      F2_mess_dauer =  F2_messzeit;             // und 1. Messung sofort abschlieen
      F2_messwert_vorhanden  = 0;               // dabei 1. Ergebnis verwerfen
      lcd_zeile1(" Frequenz in Hz "); 
      lcd_zeile2("  kein Signal   ");
      init = 1;
    } 
  
    if(F2_mess_dauer >= F2_timeout) {           // bei fehlenden Eingangsimpulsen
      set_pin(GPIOC, TRIG2, 0);                 // Trigger-Ausgang wieder auf '0'
      sende_string("kein Signal");
      init = 0;                                 // neu initialisieren
    }  

    if(F2_mess_status == MESSEN  && F2_mess_dauer >= F2_messzeit) {
      F2_mess_status = AUSLESEN;                // Zhler bei nchstem Eingangsimpuls auslesen
      set_pin(GPIOC, TRIG2, 1);                 // Trigger-Ausgang auf '1'
    }
  
    if(F2_mess_status == AUSWERTEN) {           // Ereignisse und Zeitpunkt ermittelt
      F2_mess_zeit = F2_end_zeit - F2_start_zeit; // Zeit-Differenz bilden
      F2_start_zeit = F2_end_zeit;
      
      F2_mess_ereignisse = F2_end_ereignis - F2_start_ereignis;  // Impuls-Differenz
      F2_start_ereignis = F2_end_ereignis;      // fuers naechste Intervall

      F2_mess_status = MESSEN;                  // neu starten
      F2_mess_dauer = 0;                        // und Messzeit abwarten
      if(F2_messwert_vorhanden) {               // Ergebnis berechnen und anzeigen/ausgeben
        F2_ref_frequenz = SystemTimerClock + F2_ref_offset;  // immer aktuellen Wert verwenden
        eff_zeit2 = (double)(F2_mess_zeit ) / F2_ref_frequenz;  // Zeit in Sekunden
        F2_ergebnis = ((double)F2_mess_ereignisse) / eff_zeit2; // Frequenz berechnen
        sprintf(s,"%13.7E",F2_ergebnis);        // 8-stelliges Ergebnis im Exp-Format
        lcd_zeile2(s); 
        sende_string(s);
      } else 
        F2_messwert_vorhanden = 1;              // sperre wieder aufheben
    }
}  

