/* 
Reziproke Frequenzmessung mit STM32H750 mit 10-stelliger Auflsung.
Eingangsfrequenbereich: 0,1 Hz - 100 MHz

Die Funktion der reziproken Frequenzmessung wird als bekannt vorausgesetzt.
Eine nhere Beschreibung findet sich im Programm "F2_messung.c".

Hier wird nun nicht die vorgegebene Messzeit abgewartet, um dann eine
Auswertung durchzufhren, sondern es wird nach Mglichkeit jedes einzelne
Intervall erfasst. Dazu wird die Anforderung zur Erfassung der Intervalle nicht
mehr einzeln, sondern per Timer7 ausgelst. Dieser setzt mit "F_SAMPLES" (1 MHz)
den Ausgang "TRIG2". Bei kleinerer Eingangsfrequenz wird jedes Intervall bewertet,
bei hheren Frequenzen soviele Perioden, wie in 1 s (synchronisiert) eintreffen.
Da Erfassung und Auswertung per ISR stattfinden, mu fr die restlichen 
Funktionen noch Rechenzeit brig bleiben, was hier mit dem Signal am "KONTROLL_AUSGANG"
(PB3) kontrolliert werden kann. 
Derzeit werden ca. 40% fr die ISR von Timer15 verbraucht.

Bei der Erfassung eines Intervalls werden die Summen der Einzelwerte von Ereignissen
und Zeit (reg_xi, reg_yi), die Summe aus diesen Einzelsummen (reg_sum_xi, reg_sum_yi)
und die Summen der Produkte aus reg_xi*reg_xi und reg_xi*reg_yi (reg_sum_xi2, reg_sum_xiyi)
gebildet. Ferner wird die Anzahl der Ereignisse (reg_n) gezhlt.
Nach Ablauf der Messzeit, kann aus diesen Werten mit linearer Regression die
verfeinerte Steigung der Geraden errechnet werden.

Die Grundauflsung ist 8-stellig, steigt jedoch mit Wurzel aus reg_n mit zunehmender
Eingangsfrequenz. Bei 1 s Messzeit sind folgende Auflsungen zu erwarten:

0,1 Hz - 50 Hz:    8-stellig
50 Hz - 5 kHz:     9-stellig
5 kHz - 500 kHz:  10-stellig
>= 500 kHz:       11-stellig

Erhht man die Messzeit auf 10 s erhlt man eine weitere Stelle Auflsung hinzu.
Zu beachten ist, da fr derartige Ergebnisse uerst stabile Signale fr
Eingangs- und Referenzfrequenz bentigt werden.

Die Messwerte werden ber USART3 ausgegeben: "sende_sting(s)"

2020-10-21
Michael Nowak
http://www.mino-elektronik.de
Alle Angaben ohne Gewaehr !
*/

#define F2_MESSUNG


#include <stdio.h>
#include <string.h>
#include "stm32h7xx.h"
extern void sende_string(char *s);    // Ausgabe ber USART3
extern void lcd_zeile1(char *s);
extern void lcd_zeile2(char *s);
extern uint8_t lcd_init(void);

extern uint32_t SystemTimerClock;


#define DEF_F2_TAKT       240000000   // 240 MHz
#define DEF_F2_MESSZEIT   1000        // 1 s
#define DEF_TIMEOUT       12000       // nach 12 s
#define MAX_ZEICHEN       50          // fuer Ergebnis String

#define PRIO_TIMER        4           // hohe Prioritt fr T15 und T16
#define PRIO_TIM6         14          // unkritisch
#define PRIO_TIM7         10          // < PRIO_TIMER sonst unkritisch
#define AFR_TIM15         4           // fr T15  lt. Datenblatt
#define AFR_TIM16         1           // fr T16 lt. Datenblatt

#define F_SAMPLES         1000000     // max. Messrate festlegen
#define KONTROLL_AUSGANG  3           // PB3 fr Laufzeitmessung

#define BIT(x) (1<<x)
enum gpio_modus {IN, OUT, AF, ANALOG};
enum gpio_speed {LOW_SPEED, MEDIUM_SPEED, HIGH_SPEED, V_HIGH_SPEED};
enum gpio_pullup {NO_PULL, PULL_UP, PULL_DOWN};

#define CLR_MODER         3         // Lschmaske
#define SET_GPIO          0x1       // untere Bits
#define RESET_GPIO        0x10000   // obere Bits

// PortB
#define TIM16_CAPTURE2    8
// PortC
#define TRIG2             0
// PortE
#define TIM15_ZAEHLER2    5
#define TIM15_CAPTURE2    6

enum ablauf {MESSEN=0, FERTIG};     // die Phasen der Messung

double              F2_ergebnis,                    // in Hz
                    eff_zeit2;                      // eff. Messzeit in s
double              F2_ref_frequenz = DEF_F2_TAKT;  // Vorgabewerte
int32_t             F2_ref_offset = 0,              // fr Abgleich von F2_ref
                    F2_messzeit = DEF_F2_MESSZEIT,
                    F2_timeout = DEF_TIMEOUT;

volatile uint8_t    F2_messwert_vorhanden,          // gueltige Messungen
                    F2_mess_status;                 // Ablaufsteuerung
                                   
volatile uint32_t   F2_mess_dauer;                  // in ms
uint32_t            F2_start_ereignis,              // Impulse zu Beginn der Messung
                    F2_start_zeit,                  // rel. Zeit: Beginn der Messung
                    F2_zeit_ueberlauf,
                    F2_end_zeit,                    // Zeitpunkt der Auswertung
                    F2_mess_zeit,                   // genaue Zeit der Messung
                    F2_mess_ereignisse,             // Impulse der Messung
                    F2_ueberlauf; 
volatile uint32_t   F2_end_ereignis;                // Impulse am Ende der Messung

// Variablen fr Regressionsberechnung
uint32_t  reg_n;                          // Anzahl Ereignisse bei Regression
volatile  uint32_t reg_ereignisse;        // und Endwert zur weiteren Auswertung
uint64_t  reg_xi, reg_yi;                 // summe fortlaufender Einzelwerte       
uint64_t  reg_sum_xi, reg_sum_yi;         // Summe aller Intervalle

double    reg_sum_xiyi,                   // Summe xi * yi
          reg_sum_xi2;                    // Summe xi
double    reg_xq, reg_yq;                 // Mittelwerte xi/n und yi/n

void loesche_regression(void)
{
  reg_n = 0;                              // und alle Werte der Regressionsberechnung
  reg_xi = reg_yi = 0;                    // lschen
  reg_sum_xi = reg_sum_yi = 0;
  reg_sum_xiyi = reg_sum_xi2 = 0;
  reg_xq = reg_yq = 0;
}

// direktes Setzen oder Ruecksetzen eines Port-Pins
static void set_pin(GPIO_TypeDef* GPIOx, uint8_t number, char set) {
  if(set) {
    GPIOx->BSRR = (SET_GPIO << number);       // IO-Bit setzen
  } else {
    GPIOx->BSRR = (RESET_GPIO << number);     // IO-Bit loeschen
  }
}

// 1 kHz ISR mit T6 erzeugen
void init_t6(void)
{
static uint8_t init;
  if(!init) {
    init = 1;
    RCC->APB1LENR |= RCC_APB1LENR_TIM6EN;   // Timer aktivieren
    NVIC_EnableIRQ(TIM6_DAC_IRQn);          // Int-Vector
    NVIC_SetPriority(TIM6_DAC_IRQn,PRIO_TIM6);  // geringe Prioritaet
    TIM6->PSC = SystemTimerClock/1000000-1; // Vorteiler auf 1 MHz
    TIM6->ARR = 1000-1;                     // Teiler fr 1ms
    TIM6->CR1 = TIM_CR1_CEN;                // starten
    TIM6->DIER = TIM_DIER_UIE;              // und Interrupt freigeben
  }    
} 

void TIM6_DAC_IRQHandler(void)              // Aufruf mit 1 kHz
{
  if(TIM6->SR & TIM_SR_UIF) {               // nur bei berlufen von T16
    TIM6->SR = ~TIM_SR_UIF;
    F2_mess_dauer++;                        // Messzeit erfassen
    __ISB();                                // gelschtes Flag abwarten
  }
}
 
// 1 MHz ISR mit T7 erzeugen
void init_t7(void)
{
static uint8_t init;
  if(!init) {
    init = 1;
    RCC->APB1LENR |= RCC_APB1LENR_TIM7EN;     // Timer aktivieren
    RCC->AHB4ENR |= (RCC_AHB4ENR_GPIOBEN);    // PortB aktivieren
    NVIC_EnableIRQ(TIM7_IRQn);                // Int-Vector
    NVIC_SetPriority(TIM7_IRQn,PRIO_TIM7);    // mittlere Prioritaet
    TIM7->ARR = SystemTimerClock/F_SAMPLES-1; // max. Abfragefrequenz
    TIM7->CR1 = TIM_CR1_CEN;                  // starten
    TIM7->DIER = TIM_DIER_UIE;                // und Interrupt freigeben
    GPIOB->MODER &= ~(CLR_MODER << (KONTROLL_AUSGANG * 2));// MODER PB3 = 0
    GPIOB->MODER |= (OUT << (KONTROLL_AUSGANG * 2));       // PB3 auf Ausgang         
  }
}

void TIM7_IRQHandler(void)                  // Aufruf mit F_SAMPLES
{
    TIM7->SR = 0;
    set_pin(GPIOC, TRIG2, 1);               // neue Werte anfordern
  __ISB();                                  // gelschte Flags abwarten
}

// Timer15 zhlt die ext. Eingangsimpulse: Ereignismessung
void init_t15(void)
{
uint32_t tmp;  
  RCC->APB2ENR |= RCC_APB2ENR_TIM15EN;          // Timer aktivieren
	RCC->AHB4ENR |= RCC_AHB4ENR_GPIOEEN;          // PortE aktivieren
  NVIC_EnableIRQ(TIM15_IRQn);                   // Int-Vector
  NVIC_SetPriority(TIM15_IRQn,PRIO_TIMER);      // mit hoher Prioritt

  TIM15->CR1 = 0;                               // T15 stoppen, falls schon aktiv
  tmp = GPIOE->MODER;
  tmp &= ~(CLR_MODER << (TIM15_ZAEHLER2 * 2) | 
           CLR_MODER << (TIM15_CAPTURE2 * 2));  // MODER auf 0                
  tmp |= (AF << (TIM15_ZAEHLER2 * 2) |          // altern. Funktion
          AF << (TIM15_CAPTURE2 * 2));                       
  GPIOE->MODER = tmp;

// PE5 als Zhlereingang, PE6 als capture-input zu T15
  GPIOE->AFR[0] |= (AFR_TIM15 << (TIM15_ZAEHLER2 * 4) | 
                    AFR_TIM15 << (TIM15_CAPTURE2 * 4));                   

  TIM15->CCMR1 = TIM_CCMR1_CC2S_0;              // capture zuordnung  IC2 -> TI2
  TIM15->CCER = TIM_CCER_CC2E;                  // pos. Flanken an IC2 freigeben
  TIM15->SMCR |= TIM_SMCR_TS_2 + TIM_SMCR_TS_0; // ext. Signal von TI1
  TIM15->SMCR |= TIM_SMCR_SMS_2 + TIM_SMCR_SMS_1 + 
                 TIM_SMCR_SMS_0;                // Zhlereingang von TI1

  TIM15->DIER |= TIM_DIER_CC2IE + TIM_DIER_UIE; // mit Capture + berlauf interrupts
  TIM15->CR1 |= TIM_CR1_CEN;                    // T15 starten
}

// Timer16 zhlt die internen Referenzimpulse: Zeitmessung
void init_t16(void)
{
uint32_t tmp;
  RCC->APB2ENR |= RCC_APB2ENR_TIM16EN;          // Timer aktivieren
	RCC->AHB4ENR |= RCC_AHB4ENR_GPIOBEN;          // PortB aktivieren
  NVIC_EnableIRQ(TIM16_IRQn);                   // Int-Vector
  NVIC_SetPriority(TIM16_IRQn,PRIO_TIMER);      // mit hoher Prioritt

  TIM16->CR1 = 0;                               // T16 stoppen, falls schon aktiv
  tmp = GPIOB->MODER;
  tmp &= ~(CLR_MODER << (TIM16_CAPTURE2 * 2));  // MODER PB8 = 0       
  tmp |= (2 << (TIM16_CAPTURE2 * 2));           // PB8  auf Ausgang        
  GPIOB->MODER = tmp;

// PB8 als capture-input zu T16
  GPIOB->AFR[1] |=  (AFR_TIM16 << ((TIM16_CAPTURE2-8) * 4));                   
  TIM16->CCMR1 =  TIM_CCMR1_CC1S_0;             // capture zuordnung  IC1 -> TI1
  TIM16->CCER = TIM_CCER_CC1E;                  // pos. Flanken an IC1 freigeben

  TIM16->DIER |= TIM_DIER_UIE;                  // mit interrupt bei berlauf
  TIM16->CR1 |= TIM_CR1_CEN;                    // T16 starten
}

// nur berlaufe zhlen
void TIM16_IRQHandler(void)
{
  if(TIM16->SR & TIM_SR_UIF) {                  // nur bei berlufen von T16
    TIM16->SR = ~TIM_SR_UIF;
    F2_zeit_ueberlauf += 0x10000;
  }
  __ISB();                                      // gelschtes Flag abwarten
}


// ISR zu Timer15
// synchron zu einem Eingangsimpuls Ereignisse und Zeitpunkt ermitteln
// anschlieend Auswertung mit lin. Regression
// ferner berlaufe bei der Ereigniszhlung zhlen
void TIM15_IRQHandler(void)
{
uint32_t temp;
double reg_divisor;
  if(TIM15->SR & TIM_SR_CC2IF) {                        // bei Capture-Ereignis
    set_pin(GPIOB, KONTROLL_AUSGANG, 1);                // zur Kontrolle
    TIM15->SR = ~TIM_SR_CC2IF;                          // flag wieder loeschen
    temp = TIM15->CCR2;                                 // Capture-Wert Ereignisse
    F2_end_ereignis = F2_ueberlauf + temp;              // Endwert Ereignisse
    if((TIM15->SR & TIM_SR_UIF) && (temp < 0x8000))     // evtl. Ueberlauf T15 noch offen?
      F2_end_ereignis += 0x10000;                       // nur, wenn capture-int + overflow-int gleichzeitig !

    temp = TIM16->CCR1;                                 // Capture-Wert Zeit
    F2_end_zeit = F2_zeit_ueberlauf + temp;             // Endwert Zeit
    if((TIM16->SR & TIM_SR_UIF) && (temp < 0x8000)) {   // evtl. Ueberlauf T16 noch offen?
      F2_end_zeit += 0x10000;                           // dann korrigieren
    }  

// jedes Intervall berechnen und weiterverarbeiten
    F2_mess_zeit = F2_end_zeit - F2_start_zeit;         // Zeit-Differenz bilden
    F2_start_zeit = F2_end_zeit;                        // fuers naechste Intervall
      
    F2_mess_ereignisse = F2_end_ereignis - F2_start_ereignis;  // Impuls-Differenz
    F2_start_ereignis = F2_end_ereignis;                // fuers naechste Intervall

// Anzahl und Einzelwerte fr die Regression verarbeiten      
    reg_n++;                          // anzahl der Einzelwerte
    reg_xi += F2_mess_ereignisse;     // Einzelwerte Addieren
    reg_yi += F2_mess_zeit;           // dto
    reg_sum_xi += reg_xi;             // Summe aus Summe der Einzelwerte
    reg_sum_yi += reg_yi;             // dto
    reg_sum_xiyi += reg_xi*reg_yi;    // Summe der Produkte
    reg_sum_xi2  += reg_xi*reg_xi;    // Summes der Quadrate
    
// Auswertung nach Ablauf der Messdauer starten    
    if(F2_mess_status == MESSEN &&            // wenn Auswertung aktiv
       F2_mess_dauer >= F2_messzeit) {        // und Mindestzeit erreicht
      F2_ref_frequenz = SystemTimerClock + F2_ref_offset;  // immer aktuellen Wert verwenden
      reg_xq = (double)reg_sum_xi/(double)reg_n;
      reg_yq = (double)reg_sum_yi/(double)reg_n;
      reg_divisor = (reg_sum_xi2 - reg_n * (reg_xq * reg_xq));
      if(reg_n >= 2 && reg_divisor != 0.0) {  // Regression erst ab zwei Werten          
        F2_ergebnis = F2_ref_frequenz / ((reg_sum_xiyi - reg_n * (reg_xq * reg_yq)) / reg_divisor);
      } else {                                // wenn nur ein Ereignis in reg_xi
        F2_ergebnis = ((double)reg_xi/(double)reg_yi)*F2_ref_frequenz;  // Berechnung ohne Regression
      }  
      F2_mess_status = FERTIG;                // Daten fertig fuer Anzeige
// Endwerte fr Debugger      
      reg_ereignisse = reg_n;
      F2_mess_dauer = 0;                      // neue Messzeit starten
      loesche_regression();
    }
    set_pin(GPIOC, TRIG2,0);                  // alte Anforderung lschen, TIM7 setzt neue Anforderung 
    set_pin(GPIOB, KONTROLL_AUSGANG, 0);      // zur Kontrolle
  }

// ganz zum Schluss noch die berlufe bercksichtigen
  if(TIM15->SR & TIM_SR_UIF) {                // bei berlufen von T15
     TIM15->SR = ~TIM_SR_UIF;                 // flag wieder loeschen
     F2_ueberlauf += 0x10000;                 // und berlufe zhlen
  }
  __ISB();                                    // gelschte Flags abwarten
}


/* direkte Messung mit internen Timern mit max. 10 Stellen/s
Timer15 zhlt die externen Impulse
Timer16 mit die Zeit mit internem Takt
*/
void F2_messung(uint8_t init_timer)
{
static uint8_t  init = 0;
static char s[MAX_ZEICHEN];
  if(init_timer) { 
    lcd_init();
    init = 0;                                   // Anzeige + Messung initialisieren
    init_t6();                                  // T6 fr Messdauer in 1 ms Schritten
    init_t7();                                  // TRIG2 aktivieren, max. Messrate vorgeben
    init_t15();                                 // T15 als Ereignisszhler mit capture-modus starten
    init_t16();                                 // T16 als Referenz-Zhler
    RCC->AHB4ENR |= RCC_AHB4ENR_GPIOCEN;        // PortC aktivieren
    GPIOC->MODER    &= ~(CLR_MODER << TRIG2*2); // neutral und dann
    GPIOC->MODER    |= OUT << TRIG2*2;          // als Trigger-Ausgang
    GPIOC->OSPEEDR  |= V_HIGH_SPEED << TRIG2*2; // mit steilen Flanken
  }
//  while(1) {    
    if(!init) {
      F2_mess_status = MESSEN;                  // mit MESSEN beginnen
      F2_mess_dauer =  F2_messzeit;             // und 1. Messung sofort abschlieen
      F2_messwert_vorhanden  = 0;               // dabei 1. Ergebnis verwerfen
      lcd_zeile1(" Frequenz in Hz "); 
      lcd_zeile2("  kein Signal   ");
      init = 1;
    }
    
    if(F2_mess_dauer >= F2_timeout) {
      set_pin(GPIOC, TRIG2, 0);                 // Trigger-Ausgang wieder auf '0'
      sende_string("kein Signal");
      init = 0;                                 // neu initialisieren
    }  
 
    if(F2_mess_status == FERTIG) {              // Regressionsberechnung fertig
      if(F2_messwert_vorhanden) {               // Ergebnis anzeigen/ausgeben
        sprintf(s,"%16.10E",F2_ergebnis);       // 11-stelliges Ergebnis im Exp-Format
        lcd_zeile2(s); 
        sende_string(s);
      } else {
        F2_messwert_vorhanden = 1;              // sperre wieder aufheben
      }  
      F2_mess_status  = MESSEN;                 // neue Auswertung zulassen
    }
//  }
}  

